@@ -19,11 +19,14 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
 
 // sw_main.c
-#include <stdint.h>
 #include <limits.h>
 
+#ifdef USE_SDL3
+#include <SDL3/SDL.h>
+#else
 #include <SDL2/SDL.h>
 #include <SDL2/SDL_video.h>
+#endif
 
 #include "header/local.h"
 
@@ -43,6 +46,7 @@ pixel_t		*vid_alphamap = NULL;
 light_t		vid_lightthreshold = 0;
 static int	vid_minu, vid_minv, vid_maxu, vid_maxv;
 static int	vid_zminu, vid_zminv, vid_zmaxu, vid_zmaxv;
+static qboolean IsHighDPIaware;
 
 // last position  on map
 static vec3_t	lastvieworg;
@@ -54,7 +58,6 @@ refimport_t	ri;
 
 byte		d_8to24table[256 * 4];
 
-char		skyname[MAX_QPATH];
 vec3_t		skyaxis;
 
 refdef_t	r_newrefdef;
@@ -152,6 +155,7 @@ cvar_t	*sw_texture_filtering;
 cvar_t	*r_retexturing;
 cvar_t	*r_scale8bittextures;
 cvar_t	*sw_gunzposition;
+cvar_t	*r_validation;
 static cvar_t	*sw_partialrefresh;
 
 cvar_t	*r_drawworld;
@@ -173,6 +177,8 @@ static cvar_t	*vid_fullscreen;
 static cvar_t	*vid_gamma;
 
 static cvar_t	*r_lockpvs;
+static cvar_t	*r_palettedtexture;
+cvar_t	*r_cull;
 
 // sw_vars.c
 
@@ -210,7 +216,6 @@ int		cachewidth;
 pixel_t		*d_viewbuffer;
 zvalue_t	*d_pzbuffer;
 
-static void Draw_GetPalette (void);
 static void RE_BeginFrame( float camera_separation );
 static void Draw_BuildGammaTable(void);
 static void RE_FlushFrame(int vmin, int vmax);
@@ -382,6 +387,7 @@ R_RegisterVariables (void)
 	r_retexturing = ri.Cvar_Get("r_retexturing", "1", CVAR_ARCHIVE);
 	r_scale8bittextures = ri.Cvar_Get("r_scale8bittextures", "0", CVAR_ARCHIVE);
 	sw_gunzposition = ri.Cvar_Get("sw_gunzposition", "8", CVAR_ARCHIVE);
+	r_validation = ri.Cvar_Get("r_validation", "0", CVAR_ARCHIVE);
 
 	// On MacOS texture is cleaned up after render and code have to copy a whole
 	// screen to texture, other platforms save previous texture content and can be
@@ -412,6 +418,8 @@ R_RegisterVariables (void)
 	r_customwidth = ri.Cvar_Get("r_customwidth", "1024", CVAR_ARCHIVE);
 	r_customheight = ri.Cvar_Get("r_customheight", "768", CVAR_ARCHIVE);
 	r_fixsurfsky = ri.Cvar_Get("r_fixsurfsky", "0", CVAR_ARCHIVE);
+	r_palettedtexture = ri.Cvar_Get("r_palettedtexture", "0", 0);
+	r_cull = ri.Cvar_Get("r_cull", "1", 0);
 
 	vid_fullscreen = ri.Cvar_Get( "vid_fullscreen", "0", CVAR_ARCHIVE );
 	vid_gamma = ri.Cvar_Get( "vid_gamma", "1.0", CVAR_ARCHIVE );
@@ -438,6 +446,7 @@ R_UnRegister (void)
 static void RE_ShutdownContext(void);
 static void SWimp_CreateRender(int width, int height);
 static int RE_InitContext(void *win);
+static void RE_GetDrawableSize(int* width, int* height);
 static qboolean RE_SetMode(void);
 
 /*
@@ -465,7 +474,8 @@ RE_Init(void)
 
 	r_aliasuvscale = 1.0;
 
-	Draw_GetPalette ();
+	GetPCXPalette (&vid_colormap, (unsigned *)d_8to24table);
+	vid_alphamap = vid_colormap + 64*256;
 
 	/* set our "safe" mode */
 	sw_state.prev_mode = 4;
@@ -591,10 +601,7 @@ R_ReallocateMapBuffers (void)
 
 	if (!r_numallocatedlights || r_outoflights)
 	{
-		if (!blocklights)
-		{
-			free(blocklights);
-		}
+		free(blocklights);
 
 		if (r_outoflights)
 		{
@@ -621,10 +628,7 @@ R_ReallocateMapBuffers (void)
 
 	if (!r_numallocatededges || r_outofedges)
 	{
-		if (!r_edges)
-		{
-			free(r_edges);
-		}
+		free(r_edges);
 
 		if (r_outofedges)
 		{
@@ -693,8 +697,9 @@ R_ReallocateMapBuffers (void)
 			r_outoftriangles = false;
 		}
 
-		if (r_numallocatedtriangles < vid.height)
-			r_numallocatedtriangles = vid.height;
+		// one more for the terminator
+		if (r_numallocatedtriangles < vid.height+1)
+			r_numallocatedtriangles = vid.height+1;
 
 		triangle_spans  = malloc(r_numallocatedtriangles * sizeof(spanpackage_t));
 		if (!triangle_spans)
@@ -705,7 +710,7 @@ R_ReallocateMapBuffers (void)
 		}
 		triangles_max = &triangle_spans[r_numallocatedtriangles];
 
-		R_Printf(PRINT_ALL, "Allocated %d triangles.\n", r_numallocatedtriangles);
+		R_Printf(PRINT_ALL, "Allocated %d triangle spans.\n", r_numallocatedtriangles);
 	}
 
 	if (!r_numallocatededgebasespans || r_outedgebasespans)
@@ -1005,7 +1010,7 @@ R_FindTopnode (vec3_t mins, vec3_t maxs)
 		}
 
 		splitplane = node->plane;
-		sides = BOX_ON_PLANE_SIDE(mins, maxs, (cplane_t *)splitplane);
+		sides = BOX_ON_PLANE_SIDE(mins, maxs, splitplane);
 
 		if (sides == 3)
 			return node;	// this is the splitter
@@ -1042,8 +1047,8 @@ RotatedBBox (const vec3_t mins, const vec3_t maxs, vec3_t angles, vec3_t tmins,
 
 	for (i=0 ; i<3 ; i++)
 	{
-		tmins[i] = INT_MAX; // Set maximum values for world range
-		tmaxs[i] = INT_MIN;  // Set minimal values for world range
+		tmins[i] = (vec_t)INT_MAX; // Set maximum values for world range
+		tmaxs[i] = (vec_t)INT_MIN;  // Set minimal values for world range
 	}
 
 	AngleVectors (angles, forward, right, up);
@@ -1296,6 +1301,9 @@ VectorCompareRound(const vec3_t v1, const vec3_t v2)
 	return 1;
 }
 
+cplane_t frustum[4];
+
+
 /*
 ================
 RE_RenderFrame
@@ -1322,7 +1330,7 @@ RE_RenderFrame (refdef_t *fd)
 	// compare current position with old
 	if (vid_buffer_width <= 640 ||
 	    !VectorCompareRound(fd->vieworg, lastvieworg) ||
-	    !VectorCompare(fd->viewangles, lastviewangles))
+	    !VectorCompareRound(fd->viewangles, lastviewangles))
 	{
 		fastmoving = true;
 	}
@@ -1340,6 +1348,9 @@ RE_RenderFrame (refdef_t *fd)
 
 	R_SetupFrame ();
 
+	R_SetFrustum(vup, vpn, vright, r_origin, r_newrefdef.fov_x, r_newrefdef.fov_y,
+		frustum);
+
 	// Using the current view cluster (r_viewcluster), retrieve and decompress
 	// the PVS (Potentially Visible Set)
 	R_MarkLeaves ();	// done here so we know if we're in water
@@ -1503,8 +1514,7 @@ RE_SetMode(void)
 	vid.height = r_customheight->value;
 
 	/*
-	** if this returns rserr_invalid_fullscreen then it set the mode but not as a
-	** fullscreen mode, e.g. 320x200 on a system that doesn't support that res
+	** if this returns rserr_invalid_mode then it set previous resolution
 	*/
 	if ((err = SWimp_SetMode(&vid.width, &vid.height, r_mode->value, fullscreen)) == rserr_ok)
 	{
@@ -1519,17 +1529,7 @@ RE_SetMode(void)
 	}
 	else
 	{
-		if (err == rserr_invalid_fullscreen)
-		{
-			ri.Cvar_SetValue("vid_fullscreen", 0);
-			R_Printf(PRINT_ALL, "%s() - fullscreen unavailable in this mode\n", __func__);
-
-			if (SWimp_SetMode(&vid.width, &vid.height, r_mode->value, 0) == rserr_ok)
-			{
-				return true;
-			}
-		}
-		else if (err == rserr_invalid_mode)
+		if (err == rserr_invalid_mode)
 		{
 			R_Printf(PRINT_ALL, "%s() - invalid mode\n", __func__);
 
@@ -1723,63 +1723,35 @@ RE_SetSky
 // 3dstudio environment map names
 static const char	*suf[6] = {"rt", "bk", "lf", "ft", "up", "dn"};
 static const int	r_skysideimage[6] = {5, 2, 4, 1, 0, 3};
-extern	mtexinfo_t		r_skytexinfo[6];
+extern mtexinfo_t		r_skytexinfo[6];
 
 static void
 RE_SetSky (char *name, float rotate, vec3_t axis)
 {
+	char	skyname[MAX_QPATH];
 	int		i;
-	char	pathname[MAX_QPATH];
 
 	Q_strlcpy (skyname, name, sizeof(skyname));
 	VectorCopy (axis, skyaxis);
 
 	for (i=0 ; i<6 ; i++)
 	{
-		Com_sprintf (pathname, sizeof(pathname), "env/%s%s.pcx", skyname, suf[r_skysideimage[i]]);
-		r_skytexinfo[i].image = R_FindImage (pathname, it_sky);
-		if (!r_skytexinfo[i].image)
+		image_t	*image;
+
+		image = (image_t *)GetSkyImage(skyname, suf[r_skysideimage[i]],
+			r_palettedtexture->value, (findimage_t)R_FindImage);
+
+		if (!image)
 		{
-			Com_sprintf (pathname, sizeof(pathname), "pics/Skies/%s%s.m8", skyname, suf[r_skysideimage[i]]);
-			r_skytexinfo[i].image = R_FindImage (pathname, it_sky);
+			R_Printf(PRINT_ALL, "%s: can't load %s:%s sky\n",
+				__func__, skyname, suf[r_skysideimage[i]]);
+			image = r_notexture_mip;
 		}
+
+		r_skytexinfo[i].image = image;
 	}
 }
 
-/*
-===============
-Draw_GetPalette
-===============
-*/
-static void
-Draw_GetPalette (void)
-{
-	byte	*pal, *out;
-	int		i;
-
-	/* get the palette and colormap */
-	LoadPCX ("pics/colormap.pcx", &vid_colormap, &pal, NULL, NULL);
-	if (!vid_colormap)
-		ri.Sys_Error (ERR_FATAL, "Couldn't load pics/colormap.pcx");
-	vid_alphamap = vid_colormap + 64*256;
-
-	out = d_8to24table;
-	for (i=0 ; i<256 ; i++, out+=4)
-	{
-		int r, g, b;
-
-		r = pal[i*3+0];
-		g = pal[i*3+1];
-		b = pal[i*3+2];
-
-		out[0] = r;
-		out[1] = g;
-		out[2] = b;
-	}
-
-	free (pal);
-}
-
 /*
 ===============
 RE_RegisterSkin
@@ -1841,10 +1813,19 @@ GetRefAPI(refimport_t imp)
 	// used different variable name for prevent confusion and cppcheck warnings
 	refexport_t	refexport;
 
+	// Need to communicate the SDL major version to the client.
+#ifdef USE_SDL3
+	SDL_Version ver;
+#else
+	SDL_version ver;
+#endif
+	SDL_VERSION(&ver);
+
 	memset(&refexport, 0, sizeof(refexport_t));
 	ri = imp;
 
 	refexport.api_version = API_VERSION;
+	refexport.framework_version = ver.major;
 
 	refexport.BeginRegistration = RE_BeginRegistration;
 	refexport.RegisterModel = RE_RegisterModel;
@@ -1870,6 +1851,7 @@ GetRefAPI(refimport_t imp)
 	refexport.IsVSyncActive = RE_IsVsyncActive;
 	refexport.Shutdown = RE_Shutdown;
 	refexport.InitContext = RE_InitContext;
+	refexport.GetDrawableSize = RE_GetDrawableSize;
 	refexport.ShutdownContext = RE_ShutdownContext;
 	refexport.PrepareForWindow = RE_PrepareForWindow;
 
@@ -1927,11 +1909,19 @@ RE_InitContext(void *win)
 
 	if (r_vsync->value)
 	{
+#ifdef USE_SDL3
+		renderer = SDL_CreateRenderer(window, NULL, SDL_RENDERER_PRESENTVSYNC);
+#else
 		renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC);
+#endif
 	}
 	else
 	{
+#ifdef USE_SDL3
+		renderer = SDL_CreateRenderer(window, NULL, 0);
+#else
 		renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_ACCELERATED);
+#endif
 	}
 
 	/* Select the color for drawing. It is set to black here. */
@@ -1944,8 +1934,28 @@ RE_InitContext(void *win)
 	   This will show the new, black contents of the window. */
 	SDL_RenderPresent(renderer);
 
-	vid_buffer_height = vid.height;
-	vid_buffer_width = vid.width;
+#if SDL_VERSION_ATLEAST(2, 26, 0)
+	// Figure out if we are high dpi aware.
+	int flags = SDL_GetWindowFlags(win);
+#ifdef USE_SDL3
+	IsHighDPIaware = (flags & SDL_WINDOW_HIGH_PIXEL_DENSITY) ? true : false;
+#else
+	IsHighDPIaware = (flags & SDL_WINDOW_ALLOW_HIGHDPI) ? true : false;
+#endif
+#endif
+
+	/* We can't rely on vid, because the context is created
+	   before we had a chance to overwrite it with the drawable
+	   size. */
+	if (IsHighDPIaware)
+	{
+		RE_GetDrawableSize(&vid_buffer_width, &vid_buffer_height);
+	}
+	else
+	{
+		vid_buffer_height = vid.height;
+		vid_buffer_width = vid.width;
+	}
 
 	texture = SDL_CreateTexture(renderer,
 #if SDL_BYTEORDER == SDL_BIG_ENDIAN
@@ -1962,6 +1972,19 @@ RE_InitContext(void *win)
 	return true;
 }
 
+/*
+ * Fills the actual size of the drawable into width and height.
+ */
+void RE_GetDrawableSize(int* width, int* height)
+{
+#ifdef USE_SDL3
+	SDL_GetCurrentRenderOutputSize(renderer, width, height);
+#else
+	SDL_GetRendererOutputSize(renderer, width, height);
+#endif
+}
+
+
 static void
 RE_ShutdownContext(void)
 {
@@ -2213,7 +2236,12 @@ RE_FlushFrame(int vmin, int vmax)
 
 	SDL_UnlockTexture(texture);
 
+#ifdef USE_SDL3
+	SDL_RenderTexture(renderer, texture, NULL, NULL);
+#else
 	SDL_RenderCopy(renderer, texture, NULL, NULL);
+#endif
+
 	SDL_RenderPresent(renderer);
 
 	// replace use next buffer
@@ -2341,6 +2369,48 @@ SWimp_SetMode(int *pwidth, int *pheight, int mode, int fullscreen )
 		return rserr_invalid_mode;
 	}
 
+	/* This is totaly obscure: For some strange reasons the renderer
+	   maintains three(!) repesentations of the resolution. One comes
+	   from the client and is saved in r_newrefdef. The other one
+	   is determined here and saved in vid. The third one is used by
+	   the backbuffer. Some calculations take all three representations
+	   into account.
+
+	   The values will always be the same. The GLimp_InitGraphics()
+	   call above communicates the requested resolution to the client
+	   where it ends up in the vid subsystem and the vid system writes
+	   it into r_newrefdef. The backbuffer is derived from vid.
+
+	   We can't avoid the client roundtrip, because we can get the
+	   real size of the drawable (which can differ from the resolution
+	   due to high dpi awareness) only after the render context was
+	   created by GLimp_InitGraphics() and need to communicate it
+	   somehow to the client. So we just overwrite the values saved
+	   in vid with a call to RE_GetDrawableSize(), just like the
+	   client does. This makes sure that both values are the same
+	   and everything is okay.
+
+	   We also need to take the special case fullscreen window into
+	   account. With the fullscreen windows we cannot use the
+	   drawable size, it would scale all cases to the size of the
+	   window. Instead use the drawable size when the user wants
+	   native resolution (the fullscreen window fills the screen)
+	   and use the requested resolution in all other cases. */
+	if (IsHighDPIaware)
+	{
+		if (vid_fullscreen->value != 2)
+		{
+			RE_GetDrawableSize(pwidth, pheight);
+		}
+		else
+		{
+			if (r_mode->value == -2)
+			{
+				/* User requested native resolution. */
+				RE_GetDrawableSize(pwidth, pheight);
+			}
+		}
+	}
 	return retval;
 }
 
@@ -2418,7 +2488,7 @@ SWimp_CreateRender(int width, int height)
 
 // this is only here so the functions in q_shared.c and q_shwin.c can link
 void
-Sys_Error (char *error, ...)
+Sys_Error (const char *error, ...)
 {
 	va_list		argptr;
 	char		text[4096]; // MAXPRINTMSG == 4096
@@ -2431,7 +2501,7 @@ Sys_Error (char *error, ...)
 }
 
 void
-Com_Printf(char *msg, ...)
+Com_Printf(const char *msg, ...)
 {
 	va_list	argptr;
 	va_start(argptr, msg);
